/*
** $Id: lreglib.c,v 1.38 2005/10/23 17:38:15 roberto Exp $
** Library for Register Manipulation
** See Copyright Notice in agena.h
*/


#include <stdlib.h>
#include <math.h>

#define stack_c
#define LUA_LIB

#include "agena.h"

#include "agnxlib.h"
#include "agenalib.h"
#include "agnhlps.h"  /* for introsort */
#include "lstate.h"


#define aux_checkstackindex(L, idx, procname) { \
  cell = agn_checkinteger(L, 1); \
  if (cell > -1) \
    luaL_error(L, "Error in " LUA_QS ": stack position must be negative.", "pushvalue"); \
  cell = cellstackfree(L) + cell; \
  if (cell < LUA_CELLSTACKBOTTOM) \
    luaL_error(L, "Error in " LUA_QS ": stack position is out-of-range or empty stack.", "pushvalue"); \
}
#define aux_checkstack(L, nargs, procname) { \
  if (cellstackfree(L) + (nargs) - 1 >= cellmax(L)) { \
    size_t newsize = cellmax(L) << 1; \
    L->cells[L->cellcurrent] = realloc(L->cells, newsize * sizeof(lua_Number)); \
    if (L->cells[L->cellcurrent] == NULL) \
      luaL_error(L, "Error in " LUA_QS ": memory allocation failed.", procname); \
    cellmax(L) = newsize; \
  } \
}

static int stack_pushvalued (lua_State *L) {  /* 2.9.4, push value residing at the given relative stack position to the top of the stack */
  int cell;
  aux_checkstack(L, 1, "stack.pushvalued");
  aux_checkstackindex(L, 1, "stack.pushvalued");
  cells(L, (cellstackfree(L)++)) = cells(L, cell);
  return 0;
}


static int stack_removed (lua_State *L) {  /* 2.9.4, removes the value residing at the given relative stack position,
  returns it, and moves all elements to close the space */
  int cell, i;
  lua_Number v;
  aux_checkstackindex(L, 1, "stack.removed");
  v = cells(L, cell);
  for (i=cell + 1; i < cellstackfree(L); i++)
    cells(L, i - 1) = cells(L, i);
  cellstackfree(L)--;
  lua_pushnumber(L, v);
  return 1;
}


static int stack_insertd (lua_State *L) {  /* 2.9.4, inserts an element to the given stack position shifting up other elements to open space */
  int cell, i;
  aux_checkstack(L, 1, "stack.insertd");
  aux_checkstackindex(L, 1, "stack.insertd");
  for (i=(cellstackfree(L)++); i > cell; i--)
    cells(L, i) = cells(L, i - 1);
  cells(L, cell) = agn_checknumber(L, 2);
  return 0;
}


static int stack_dumpd (lua_State *L) {  /* 2.9.4, returns all or a given number of values in the built-in stack
  in a new sequence and pops them from the stack. The number of pre-allocated slots is not changed, see shrinkd. */
  int i, n, c, csf;  // cn, oldcellcurrent;
  //oldcellcurrent = L->cellcurrent;    /* 2.9.7 */
  //cn = agnL_optinteger(L, 2, 1);
  //if (cn < 1 || cn > 3)
  //  luaL_error(L, "Error in " LUA_QS ": invalid argument %d.", "stack.dumpd", cn);
  //L->cellcurrent = cn - 1;
  csf = cellstackfree(L);
  n = luaL_optinteger(L, 1, csf - LUA_CELLSTACKBOTTOM);  /* 2.9.5 */
  if (n < 0 || n > csf - LUA_CELLSTACKBOTTOM)
    luaL_error(L, "Error in " LUA_QS ": invalid argument.", "stack.dumpd");
  agn_createseq(L, n);
  c = 0;
  for (i=csf - n; i < csf; i++)
    lua_seqsetinumber(L, -1, ++c, cells(L, i));
  cellstackfree(L) -= n;
  //L->cellcurrent = oldcellcurrent;
  return 1;
}


static int stack_sized (lua_State *L) {  /* 2.9.4, returns the current number of elements in the built-in current stack
  along with the current maximum number of slots allocated by the system. See also: allotted. */
  lua_pushinteger(L, cellstackfree(L));
  lua_pushinteger(L, cellmax(L));
  return 2;
}


static int stack_shrinkd (lua_State *L) {  /* 2.9.4; if possible, shrinks the number of pre-allocated but not assigned slots in the
  built-in stack. The function returns the new number of pre-allocated slots but does not pop any elements. The function is useful
  to reduce memory consumption if a lot of numbers have been removed from the current stack. */
  size_t p, csf, cn, i, nargs, oldcellnumber;
  nargs = lua_gettop(L);
  if (nargs == 0) {
    cn = L->cellcurrent;
    csf = cellstackfree(L);
    p = 2 << ( ( (csf < LUA_CELLSTACKSIZE) ? ceillog2(LUA_CELLSTACKSIZE) : ceillog2(csf) ) - 1 );
    L->cells[cn] = realloc(L->cells[cn], p * sizeof(lua_Number));
    cellmax(L) = p;
    lua_pushinteger(L, p);
    i = 1;
  } else {  /* 2.9.7: shrink all stacks */
    oldcellnumber = L->cellcurrent;
    for (i=0; i < 3; i++) {
      L->cellcurrent = cn = i;
      csf = cellstackfree(L);
      p = 2 << ( ( (csf < LUA_CELLSTACKSIZE) ? ceillog2(LUA_CELLSTACKSIZE) : ceillog2(csf) ) - 1 );
      L->cells[cn] = realloc(L->cells[cn], p * sizeof(lua_Number));
      cellmax(L) = p;
      lua_pushinteger(L, p);
    }
    i--;
    L->cellcurrent = oldcellnumber;
  }
  return i;
}


/* Reverses the positions of the elements in the built-in stack. The function returns nothing. */
static int stack_reversed (lua_State *L) {  /* 2.9.4 */
  size_t i, offset, csf;
  int n;  /* 2.9.7 fix */
  lua_Number t;
  csf = cellstackfree(L);
  n = luaL_optinteger(L, 1, csf - LUA_CELLSTACKBOTTOM);  /* 2.9.6 */
  if (n < 0 || n > csf - LUA_CELLSTACKBOTTOM)
    luaL_error(L, "Error in " LUA_QS ": invalid argument.", "stack.reversed");
  offset = csf - n;
  n = csf;
  for (i=0 + offset; i < (n + offset)/2; i++) {
    t = cells(L, i);
    cells(L, i) = cells(L, n - i - 1 + offset);
    cells(L, n - i - 1 + offset) = t;
  }
  return 0;
}


/* Moves all the elements in the built-in stack n places from the bottom to the top if n is positive,
   and n places from the top to the bottom if n is negative. The function returns nothing. */
static int stack_rotated (lua_State *L) {  /* 2.9.4 */
  size_t s, i, j;
  int n;
  lua_Number t;
  s = cellstackfree(L);
  n = agnL_optinteger(L, 1, 1);
  if (s < 2) return 0;
  if (n < 0) {
    for (j=0; j < -n; j++) {
      t = cells(L, 0);
      for (i=0; i < s - 1; i++) cells(L, i) = cells(L, i + 1);
      cells(L, s - 1) = t;
    }
  } else if (n > 0) {
    for (j=0; j < n; j++) {
      t = cells(L, s - 1);
      for (i=s - 1; i > 0; i--) cells(L, i) = cells(L, i - 1);
      cells(L, 0) = t;
    }
  }
  return 0;
}


/* Sorts the last given number of elements pushed onto the current stack in ascending order using the fast Introsort
   algorithm. If no argument is passed, all elements in the current stack are processed. The function returns nothing. */
static int stack_sorted (lua_State *L) {  /* 2.9.7 */
  size_t csf;
  int n;
  csf = cellstackfree(L);
  n = luaL_optinteger(L, 1, csf - LUA_CELLSTACKBOTTOM);
  if (n < 0 || n > csf - LUA_CELLSTACKBOTTOM)
    luaL_error(L, "Error in " LUA_QS ": invalid argument.", "stack.sorted");
  tools_dintrosort(L->cells[L->cellcurrent], csf - n, csf - 1, 0, 2*log(csf)/log(2));
  return 0;
}


static int stack_resetd (lua_State *L) {  /* 2.9.4, Clears the entire stack so that it becomes empty. The functions
  does not return anything and should be used if succeedingly working with stacks in a session. The number of
  pre-allocated slots is not changed, however, see shrinkd. */
  size_t nargs;
  nargs = lua_gettop(L);
  if (nargs == 0) {
    cellstackfree(L) = LUA_CELLSTACKBOTTOM;
  } else {
    size_t i, cnr, oldcnr;
    oldcnr = L->cellcurrent;
    for (i=1; i <= nargs; i++) {
      cnr = agn_checkinteger(L, i);
      if (cnr < 1 || cnr > 3)
        luaL_error(L, "Error in " LUA_QS ": invalid stack number %d.", "stack.resetd", cnr);
      L->cellcurrent = cnr - 1;
      cellstackfree(L) = LUA_CELLSTACKBOTTOM;
    }
    L->cellcurrent = oldcnr;
  }
  return 0;
}


static int stack_selected (lua_State *L) {  /* 2.9.7, returns the current stack number. */
  lua_pushinteger(L, L->cellcurrent + 1);
  return 1;
}


static int stack_choosed (lua_State *L) {  /* 2.9.7, Returns the number of the stack with the least number of stored values. */
  int i, min, x, smallest, oldstack;
  min = smallest = MAX_INT;
  oldstack = L->cellcurrent;
  for (i=2; i >= 0 ; i--) {
    L->cellcurrent = i;
    x = cellstackfree(L);
    if (x <= min) { min = x; smallest = i; }  /* always return the smallest stack number */
  }
  L->cellcurrent = oldstack;
  lua_pushinteger(L, smallest + 1);
  return 1;
}


/* }====================================================== */

static const luaL_Reg stack_funcs[] = {
  {"choosed", stack_choosed},               /* added January 13, 2016 */
  {"dumpd", stack_dumpd},                   /* added November 29, 2015 */
  {"insertd", stack_insertd},               /* added November 29, 2015 */
  {"pushvalued", stack_pushvalued},         /* added November 29, 2015 */
  {"removed", stack_removed},               /* added November 29, 2015 */
  {"resetd", stack_resetd},                 /* added November 30, 2015 */
  {"reversed", stack_reversed},             /* added November 30, 2015 */
  {"rotated", stack_rotated},               /* added November 30, 2015 */
  {"selected", stack_selected},             /* added January 11, 2016 */
  {"shrinkd", stack_shrinkd},               /* added November 30, 2015 */
  {"sized", stack_sized},                   /* added November 30, 2015 */
  {"sorted", stack_sorted},                 /* added January 03, 2016 */
  {NULL, NULL}
};


LUALIB_API int luaopen_stack (lua_State *L) {
  luaL_register(L, AGENA_STACKLIBNAME, stack_funcs);
  return 1;
}

